#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Capability testing utilities """

# pylint: disable=C0301, E0401, R0903, R0913

from datetime import datetime
from typing import Any, Dict, List, Optional
from unittest.mock import MagicMock

import pendulum
import pytest
from pyramid.config import Configurator
from pyramid.testing import DummyRequest, setUp, tearDown

from workspaces.capability.enums import CapabilityRequestState
from workspaces.capability.helpers import Parameter
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.products.schema_interfaces import FutureProductIF


@pytest.fixture(scope="module")
def test_config() -> Configurator:
    """
    Returns a dummy Configurator object for testing purposes with set up and teardown

    :return: Dummy Configurator
    """
    request = DummyRequest()
    config = setUp(request=request)
    config.add_request_method(
        lambda lookup: MagicMock(),
        "capability_info",
        reify=True,
    )
    yield config
    tearDown()


class MockCapabilityInfo(MagicMock):
    """Not a CapabilityInfo, but plays one on TV."""

    capabilities = [
        Capability(name="null", max_jobs=2, enabled=True, has_image_products=False, paused=False),
        Capability(name="error", max_jobs=-1, enabled=True, has_image_products=False),
        Capability(
            name="std_restore_imaging",
            max_jobs=1,
            enabled=True,
            has_image_products=True,
        ),
    ]
    capability_requests = [
        CapabilityRequest(
            id=0,
            state="Created",
            capability_name="null",
            versions=[],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        ),
        CapabilityRequest(
            id=1,
            state="Complete",
            capability_name="null",
            versions=[CapabilityVersion(capability_request_id=1, version_number=1, parameters='{"arguments": "-pe"}')],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        ),
        CapabilityRequest(
            id=2,
            state="Created",
            capability_name="test",
            versions=[],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        ),
        CapabilityRequest(
            id=3,
            state="Submitted",
            capability_name="null",
            versions=[CapabilityVersion(capability_request_id=1, version_number=1, parameters='{"arguments": "-g"}')],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        ),
    ]

    carta_capability_version = CapabilityVersion(
        capability_request_id=21,
        version_number=2,
        parameters={
            "product_locator": "uid://evla/execblock/35b23178-0dde-4167-82fd-9bd6128a80a1",
            "cal_locator": "uid://evla/calibration/6a940d35-c714-47a2-ac4a-1df6182ee83f",
            "user_email": "jgoldste@nrao.edu",
        },
        files=[],
        workflow_metadata={"carta_url": "https://pr-dsoc-dev.nrao.edu/pKP8pcGvI02M/"},
    )
    capability_requests.append(
        CapabilityRequest(
            id=carta_capability_version.capability_request_id,
            state="Complete",
            capability_name="std_restore_imaging",
            versions=[
                CapabilityVersion(capability_request_id=1, version_number=1, parameters='{"arguments": "-g"}'),
                carta_capability_version,
            ],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        )
    )

    capability_executions = []
    capability_versions = [carta_capability_version]

    def edit_capability(
        self,
        name: str,
        steps: str = None,
        max_jobs: int = None,
        enabled: bool = None,
        has_image_products: bool = None,
    ) -> bool:
        """
        Change this capability according to params passed in.

        :param name: capability name
        :param steps: capability steps, if changing them
        :param max_jobs: maximum number of jobs, if changing
        :param enabled: enable or disable the capability

        :return:
        """
        if name == "error":
            # This is here to mimic the case where an update fails to happen
            return False
        for capability in self.capabilities:
            if name == capability.name:
                if steps:
                    capability.steps = steps
                if max_jobs:
                    capability.max_jobs = max_jobs
                if enabled:
                    capability.enabled = enabled
                if has_image_products:
                    capability.has_image_products = has_image_products
                return True
        return False

    def lookup_capability(self, capability_name: str) -> Capability:
        """
        Find the capability with this name.

        :param capability_name:
        :return:
        """
        for capability in self.capabilities:
            if capability_name == capability.name:
                return capability
        return None

    def lookup_capability_request(self, request_id: int) -> CapabilityRequest:
        """
        Find the capability request with this request ID.

        :param request_id:
        :return:
        """
        for capability_request in self.capability_requests:
            if request_id == capability_request.id:
                return capability_request
        return None

    def lookup_capability_version(self, request_id: int, version_number: int) -> CapabilityVersion:
        """
        Find the capability version with this request ID and version number.

        :param request_id: the capability request ID
        :return:
        """
        for capability_request in self.capability_requests:
            current_version = capability_request.current_version
            if current_version.capability_request_id == request_id and current_version.version_number == version_number:
                return current_version

        return None

    def lookup_active_capability_requests(self, capability_name: str) -> Optional[Dict[str, List[Dict]]]:
        if capability_name not in [capability.name for capability in self.capabilities]:
            return None

        active_requests = []

        for capability_request in self.capability_requests:
            if capability_request.capability_name == capability_name and capability_request.state != "Complete":
                active_requests.append(capability_request.__json__())

        return {"active_requests": active_requests}

    def lookup_created_capability_requests(self, capability_name: str) -> Optional[Dict[str, List[Dict]]]:
        if capability_name not in [capability.name for capability in self.capabilities]:
            return None

        active_requests = []

        for capability_request in self.capability_requests:
            if capability_request.capability_name == capability_name and capability_request.state == "Created":
                active_requests.append(capability_request.__json__())

        return {"active_requests": active_requests}

    def delete_capability_request(self, request_id: int) -> int:
        """
        Delete the capability request with this ID.

        :param request_id:
        :return:
        """
        for capability_request in self.capability_requests:
            if request_id == capability_request.id:
                if self.can_delete_capability_request(request_id):
                    self.capability_requests.remove(capability_request)
                    return 1
                return -1

        return 0

    def can_delete_capability_request(self, request_id) -> bool:
        """
        May I delete this capability request, pretty please?

        :param request_id: the ID of the doomed request

        :return:
        """
        for execution in self.capability_executions:
            if execution.capability_request_id == request_id:
                return False
        return True

    def save_entity(self, entity: Any):
        """
        Save this entity, whatever it may be

        :param entity: Capability, CapabilityRequest, or CapabilityExecution
        :return:
        """
        # don't append it if we already know the ID
        if hasattr(entity, "id") and entity.id is not None:
            return

        if isinstance(entity, Capability):
            self.capabilities.append(entity)
        elif isinstance(entity, CapabilityRequest):
            entity.id = len(self.capability_requests)
            self.capability_requests.append(entity)
        elif isinstance(entity, CapabilityExecution):
            entity.id = len(self.capability_executions)
            self.capability_executions.append(entity)
        elif isinstance(entity, CapabilityVersion):
            entity.id = len(self.capability_versions)
            self.capability_versions.append(entity)

    def create_capability_request(
        self,
        capability_name,
        parameters: List[ParameterIF] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequest:
        capability = self.lookup_capability(capability_name)
        request = CapabilityRequest(
            state=CapabilityRequestState.Created.name,
            capability=capability,
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=parameters)],
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )
        self.save_entity(request)
        return request

    def create_capability_version(
        self,
        capability_request_id: int,
        parameters: List[ParameterIF] = None,
    ) -> CapabilityVersion:
        request = self.lookup_capability_request(capability_request_id)
        current_version_files = request.current_version.files
        version = CapabilityVersion(
            capability_request_id=capability_request_id,
            version_number=len(request.versions) + 1,
            parameters=parameters,
            request=request,
        )
        self.save_entity(version)

        # Reset request state to Created
        request.state = CapabilityRequestState.Created.name
        self.save_entity(request)

        self.copy_previous_version_files(version=version, previous_version_files=current_version_files)
        return version


class MockCapabilityLauncher(MagicMock):
    """Not a capability launcher, but plays one on TV"""

    def launch_capability(self, request: CapabilityRequest) -> CapabilityExecution:
        return CapabilityExecution(created_at=datetime.now(), updated_at=datetime.now())


class MockCapabilityService(MagicMock):
    """Not a capabilitiy service, but wishes it were"""

    def __init__(self, capability_info: MockCapabilityInfo):
        super().__init__()
        self.capability_info = capability_info
        self.execution_manager = MockExecutionManager(capability_info)

    def create_request(
        self,
        capability_name: str,
        parameters: List[Parameter] = None,
    ) -> CapabilityRequest:
        """
        Mocked version of the corresponding `CapabilityService.create_request` method

        :param capability_name: Name of capability
        :param parameters: Parameters for request

        :return: Newly created request
        """
        capability = self.capability_info.lookup_capability(capability_name)

        request = CapabilityRequest(
            state=CapabilityRequestState.Created.name,
            capability=capability,
            capability_name=capability_name,
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=str(parameters))],
            created_at=pendulum.now(),
            updated_at=pendulum.now(),
        )
        self.capability_info.save_entity(request)
        return request

    def create_new_version(self, capability_request_id: int, parameters: List[Parameter] = None) -> CapabilityVersion:
        """
        Mocked version of the corresponding `CapabilityService.create_new_version` method

        :param capability_request_id: id of parent capability request
        :param parameters: Parameters for request
        :return: Newly created version
        """
        request = self.capability_info.lookup_capability_request(capability_request_id)
        version = CapabilityVersion(
            capability_request_id=capability_request_id,
            version_number=len(request.versions) + 1,
            parameters=parameters,
            request=request,
        )
        self.capability_info.save_entity(version)
        return version

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecution:
        """
        Run this capability request.

        :param request:
        :return:
        """
        return self.execution_manager.run_capability(request)


class MockExecutionManager:
    """So we don't have to use a real ExecutionManager."""

    def __init__(self, capability_info: MockCapabilityInfo):
        self.capability_info = capability_info

    @pytest.mark.skip(reason="Broken by state machine rework")
    def run_capability(self, capability_request: CapabilityRequest) -> CapabilityExecution:
        """
        Mock run_capability method

        :param capability_request: Request to make an execution for
        :return: Mocked execution
        """
        execution = CapabilityExecution(
            state="Start",
            capability_request_id=capability_request.id,
            version=capability_request.current_version,
            created_at=datetime.now(),
            updated_at=datetime.now(),
        )
        self.capability_info.save_entity(execution)
        return execution


@pytest.fixture(scope="function")
def request_null_capability() -> DummyRequest:
    """
    Returns a dummy request object with a mocked capability_info

    :return: DummyRequest with configured mock CapabilityInfo
    """

    mock_capability_info = MockCapabilityInfo()
    request = DummyRequest(
        capability_info=mock_capability_info,
        capability_service=MockCapabilityService(mock_capability_info),
        capability_launcher=MockCapabilityLauncher(),
    )
    return request
